@@ -6,7 +6,6 @@ import json
 from http import HTTPStatus
 from typing import Dict, List
 
-import numpy as np
 import pandas as pd
 import requests
 from requests.exceptions import HTTPError, JSONDecodeError
@@ -19,6 +18,121 @@ API_URL = api_config.url
 API_VERSION = 2
 
 
+def get_intraday_trades(
+    price_areas: List[str] = None,
+    start: str = None,
+    end: str = None,
+    *,
+    access_token: str = None,
+):
+    """Get raw intraday trades from Archimedes Database
+
+    This function can be used to fetch raw time series from the Archimedes Database without any post-processing.
+    To see which series are available, use `list_ids()`.
+
+    Example:
+        >>> archimedes.get(
+        >>>     price_areas=["NO1",],
+        >>>     start="2020-06-20T04:00:00+00:00",
+        >>>     end="2020-06-20T09:00:00+00:00",
+        >>> )
+                from_dt to_dt   series_id   version price   trade_time  buy_area    sell_area   attributes
+        0	2022-03-01T00:00:00+00:00	2022-03-01T01:00:00+00:00	NP/IntradayTrades	1646193645	146.63	2022-02-28T17:02:06.934+0000	OPX	NO1	{'price': 146.63,...,  'product_code': 'PH-20220301-01'}
+        1	2022-03-01T00:00:00+00:00	2022-03-01T01:00:00+00:00	NP/IntradayTrades	1646193645	146.63	2022-02-28T17:02:06.934+0000	OPX	NO1	{'price': 146.63,... , 'product_code': 'PH-20220301-01'}
+        2	2022-03-01T00:00:00+00:00	2022-03-01T01:00:00+00:00	NP/IntradayTrades	1646193645	146.63	2022-02-28T17:02:06.934+0000	OPX	NO1	{'price': 146.63, ..., 'product_code': 'PH-20220301-01'}
+        ...
+        155	2022-03-01T23:00:00+00:00	2022-03-02T00:00:00+00:00	NP/IntradayTrades	1646193645	148.99	2022-03-01T13:05:40.934+0000	NL	NO1	{'price': 148.99,... , 'product_code': 'PH-20220301-24'}
+        156	2022-03-01T23:00:00+00:00	2022-03-02T00:00:00+00:00	NP/IntradayTrades	1646193645	148.8	2022-03-01T12:57:58.777+0000	NO1	FI	{'price': 148.8,...,  'product_code': 'PH-20220301-24'}
+        157	2022-03-01T23:00:00+00:00	2022-03-02T00:00:00+00:00	NP/IntradayTrades	1646193645	148.8	2022-03-01T12:58:02.115+0000	NO1	FI	{'price': 148.8, ... , 'product_code': 'PH-20220301-24'}
+
+
+    Args:
+        price_areas (List[str], optional): The price areas to pick, all price areas if None. Defaults to None.
+        start (str, optional): The first datetime to fetch (inclusive). Returns all if None. Defaults to None.
+        end (str, optional): The last datetime to fetch (exclusive). Returns all if None. Defaults to None.
+        access_token (str, optional): None - access token for the API
+
+    Returns:
+        DataFrame with all the time series data
+
+    Raises:
+        HTTPError: If an http error occurs when requesting the API.
+        NoneAuth: If the user is unauthorized or if the authorization has expired.
+    """
+
+    if isinstance(price_areas, str):
+        price_areas = [price_areas]
+
+    if start is None:
+        start = archimedes.constants.DATE_LOW
+    else:
+        start = pd.to_datetime(start)
+
+    if end is None:
+        end = archimedes.constants.DATE_HIGH
+    else:
+        end = pd.to_datetime(end)
+
+    query = {
+        "series_id": "NP/IntradayTrades",
+        "price_areas": price_areas,
+        "start": start,
+        "end": end,
+        "flatten_columns": False,
+    }
+
+    observation_data = _make_api_request(
+        f"{API_URL}/v{API_VERSION}/observation_json/get",
+        access_token=access_token,
+        params=query,
+    )
+
+    if not len(observation_data):
+        return pd.DataFrame(
+            columns=[
+                "from_dt",
+                "to_dt",
+                "series_id",
+                "version",
+                "price",
+                "trade_time",
+                "buy_area",
+                "sell_area",
+                "attributes",
+            ]
+        )
+    observation_data = [
+        {
+            **i,
+            "price": i["value"].get("price"),
+            "trade_time": i["value"].get("trade_time"),
+        }
+        for i in observation_data
+    ]
+    observation_data = pd.DataFrame.from_dict(observation_data)
+
+    # Extracting price area and series id
+    observation_data[
+        ["series_id1", "series_id2", "buy_area", "sell_area"]
+    ] = observation_data["series_id"].str.split("/", 3, expand=True)
+    observation_data["series_id"] = (
+        observation_data["series_id1"] + "/" + observation_data["series_id2"]
+    )
+    observation_data["attributes"] = observation_data["value"]
+    observation_data["attributes_str"] = observation_data["value"].astype(str)
+    observation_data = observation_data.drop_duplicates(
+        ["from_dt", "series_id", "attributes_str"]
+    )
+    observation_data = observation_data.drop(
+        ["series_id1", "series_id2", "value", "version", "attributes_str"], axis=1
+    )
+    observation_data = observation_data.sort_values(by=["from_dt"]).reset_index(
+        drop=True
+    )
+
+    return observation_data
+
+
 def get(
     series_ids: List[str],
     price_areas: List[str] = None,
@@ -87,20 +201,19 @@ def get(
         "flatten_columns": True,
     }
 
-    data = _make_api_request(
+    df = _make_api_request(
         f"{API_URL}/v{API_VERSION}/data/get",
         access_token=access_token,
         params=query,
     )
 
-    df = pd.DataFrame.from_dict(data)
-
-    row_count = df.shape[0]
-    if row_count == 0:
-        return df
-
-    df = df.sort_values(by=["from_dt", "version"])
-
+    if len(df):
+        df = pd.DataFrame.from_dict(df)
+        df = df.sort_values(by=["from_dt", "version"])
+    else:
+        df = pd.DataFrame(
+            columns=["series_id", "from_dt", "ref_dt", "value", "version", "price_area"]
+        )
     df = df.pivot_table(
         values="value",
         columns=["series_id", "price_area"],
@@ -108,8 +221,51 @@ def get(
         aggfunc="last",
     )
 
-    df.index = pd.to_datetime(df.index)
+    del query["series_ids"]
+    for series_id in series_ids:
+        query["series_id"] = series_id
+        observation_data = _make_api_request(
+            f"{API_URL}/v{API_VERSION}/observation_json/get",
+            access_token=access_token,
+            params=query,
+        )
+        if len(observation_data):
+            observation_data = pd.DataFrame.from_dict(observation_data)
 
+            observation_data[
+                ["series_id1", "series_id2", "price_area"]
+            ] = observation_data["series_id"].str.split("/", 2, expand=True)
+            observation_data["series_id"] = (
+                observation_data["series_id1"] + "/" + observation_data["series_id2"]
+            )
+            observation_data = observation_data.drop(
+                ["series_id1", "series_id2"], axis=1
+            )
+
+            if series_id == "NP/IntradayTrades":
+
+                def agg_func(x):
+                    return list(x)
+
+                sort_headers = ["from_dt"]
+            else:
+                agg_func = "last"
+                sort_headers = ["from_dt", "version"]
+
+            observation_data = observation_data.sort_values(by=sort_headers)
+            observation_data = observation_data.pivot_table(
+                values="value",
+                columns=["series_id", "price_area"],
+                index="from_dt",
+                aggfunc=agg_func,
+            )
+
+            df = pd.merge(
+                df, observation_data, left_index=True, right_index=True, how="outer"
+            )
+
+    df.index = pd.to_datetime(df.index)
+    df = df.fillna("")
     return df
 
 
@@ -177,7 +333,7 @@ def get_predictions(
     end: str = None,
     *,
     access_token: str = None,
-) -> List:
+) -> pd.DataFrame:
     """Get any number of predictions
 
     This function can be used to fetch predictions from the Archimedes Database.
@@ -262,10 +418,21 @@ def list_series_price_areas(series_id: str, *, access_token: str = None):
         "series_id": series_id,
     }
     data = _make_api_request(
-        f"{API_URL}/v{API_VERSION}/data/list_series_price_areas", access_token=access_token, params=query
+        f"{API_URL}/v{API_VERSION}/data/list_series_price_areas",
+        access_token=access_token,
+        params=query,
     )
+    data = pd.DataFrame.from_dict(data)
 
-    price_area_df = pd.DataFrame.from_dict(data)
+    observation_data = _make_api_request(
+        f"{API_URL}/v{API_VERSION}/observation_json/list_series_price_areas",
+        access_token=access_token,
+        params=query,
+    )
+    observation_data = pd.DataFrame.from_dict(observation_data)
+
+    price_area_df = pd.concat([data, observation_data]).drop_duplicates()
+    price_area_df = price_area_df.sort_values("price_areas").reset_index(drop=True)
     return price_area_df
 
 
@@ -289,34 +456,28 @@ def list_ids(sort: bool = False, *, access_token: str = None):
         DataFrame with all available list_ids
     """
     data = _make_api_request(
-        f"{API_URL}/v{API_VERSION}/data/list_ids",
-        access_token=access_token
+        f"{API_URL}/v{API_VERSION}/data/list_ids", access_token=access_token
     )
+    data = pd.DataFrame.from_dict(data)
 
-    series_df = pd.DataFrame.from_dict(data)
+    observation_data = _make_api_request(
+        f"{API_URL}/v{API_VERSION}/observation_json/list_ids", access_token=access_token
+    )
+    observation_data = pd.DataFrame.from_dict(observation_data)
 
+    series_df = pd.concat([data, observation_data]).drop_duplicates()
+    series_df = series_df.sort_values(["series_id"]).reset_index(drop=True)
     if not sort:
         return series_df
 
-    # get list of data-sources and sort alphabetically
-    data_sources = np.unique([ids.split("/")[0] for ids in series_df.values.flatten()])
-    data_sources = np.sort(data_sources)
+    series_df["pre"] = series_df["series_id"].str.split("/", 1).str[0]
+    series_df = pd.DataFrame.from_dict(
+        series_df.groupby("pre")["series_id"].apply(list).to_dict(), orient="index"
+    ).transpose()
+    series_df = series_df[sorted(series_df.columns)]
 
-    sorted_df = pd.DataFrame()
-    for source in data_sources:
-        source_ids = series_df[
-            pd.Series(series_df.values.flatten()).str.startswith(source)
-        ]
-        # sort series_ids alphabetically and reset dataframe index
-        source_ids = source_ids.sort_values(by="series_id")
-        source_ids = source_ids.reset_index(drop=True)
-        # append column of series_ids to frame
-        sorted_df = pd.concat([sorted_df, source_ids], axis=1)
-
-    # set name of columns and replace nans with empty strings
-    sorted_df.columns = data_sources
-    sorted_df = sorted_df.fillna("")
-    return sorted_df.copy()
+    series_df = series_df.fillna("")
+    return series_df.copy()
 
 
 def list_prediction_ids(*, access_token: str = None):
@@ -332,7 +493,9 @@ def list_prediction_ids(*, access_token: str = None):
         23                         PX/rk-naive
     """
 
-    data = _make_api_request(f"{API_URL}/v{API_VERSION}/data/list_prediction_ids", access_token=access_token)
+    data = _make_api_request(
+        f"{API_URL}/v{API_VERSION}/data/list_prediction_ids", access_token=access_token
+    )
 
     return pd.DataFrame.from_dict(data)
 
@@ -367,7 +530,9 @@ def get_predictions_ref_dts(prediction_id: str = None, *, access_token: str = No
 def _make_api_request(url, method="GET", access_token=None, *args, **kwargs):
     if access_token is None:
         if archimedes_auth is None:
-            raise NoneAuth("access_token parameter must be passed when using USE_WEB_AUTHENTICATION")
+            raise NoneAuth(
+                "access_token parameter must be passed when using USE_WEB_AUTHENTICATION"
+            )
         access_token = archimedes_auth.get_access_token_silent()
 
     headers = kwargs.pop("headers", None)
@@ -489,7 +654,9 @@ def forecast_list_ref_times(
     }
 
     data = _make_api_request(
-        f"{API_URL}/v{API_VERSION}/forecast/list_ref_times", access_token=access_token, params=query
+        f"{API_URL}/v{API_VERSION}/forecast/list_ref_times",
+        access_token=access_token,
+        params=query,
     )
 
     return pd.DataFrame.from_dict(data)
@@ -543,7 +710,11 @@ def forecast_diff(
         "ref_time2": ref_time2,
     }
 
-    data = _make_api_request(f"{API_URL}/v{API_VERSION}/forecast/diff", access_token=access_token, params=query)
+    data = _make_api_request(
+        f"{API_URL}/v{API_VERSION}/forecast/diff",
+        access_token=access_token,
+        params=query,
+    )
 
     return pd.DataFrame.from_dict(data)
 
@@ -594,7 +765,9 @@ def forecast_get_by_ref_time_interval(
     }
 
     data = _make_api_request(
-        f"{API_URL}/v{API_VERSION}/forecast/get_by_ref_time_interval", access_token=access_token, params=query
+        f"{API_URL}/v{API_VERSION}/forecast/get_by_ref_time_interval",
+        access_token=access_token,
+        params=query,
     )
 
     return pd.DataFrame.from_dict(data)
